fM3G = getOpenFileName "Choose Model File:" types:"Real Racing 3 Vehicle Meshes (*.m3g)|*.m3g|"

fn readStr binstr strLen =
(
	str = ""
	for w=1 to strLen do(str += (bit.intasChar(readByte binstr #unsigned)))
	return str
)

if fM3G != undefined do
(
	f = fopen fM3G "rb"
	clearlistener()

	fileIdentifier = readStr f 12	
	if fileIdentifier != "JSR184\r\n\x1A\n" then
	(
		messageBox "Incorrect file identifier. Ending script."
		fclose f
	)
	else
	(
		max modify mode
		cui.expertModeOn()

		objDataArr = #()
		globObjCount = 0

		endOfFile = getFileSize fM3G

		mArr=#()
		while ftell f < endOfFile do
		(
			compressionScheme  = readbyte f
			sectionTotalLength = readlong f
			sectionUncompressedLength = readlong f
			sectionUncompressedLength += ftell f
			while ftell f < sectionUncompressedLength do
			(
				objectType   = readbyte f
				objectLength = readlong f

				globObjCount += 1
				objDataArr[globObjCount]=#() -- create an array for the object to store data in
				case objectType of
				(
					20: -- vertex array
					(
						fseek f 12#seek_cur
						bytes  = readbyte f #unsigned
						fields = readbyte f #unsigned
						compr  = readbyte f #unsigned
						count  = readshort f #unsigned

						case (bytes+fields) of
						(
							4: -- UVs (maybe colors too?)
							(
								for x = 1 to count do
								(
									tu=readshort f
									tv=readshort f
									append (objDataArr[globObjCount]) ([tu,-tv,0]/2048.0)
								)
							)
							5: -- normals
							(
								for x = 1 to count do
								(
									nx=readshort f
									ny=readshort f
									nz=readshort f
									append (objDataArr[globObjCount]) ([-nx,nz,ny]/32767.0)
								)
							)
							7: -- vertices
							(
								for x = 1 to count do
								(
									vx=readfloat f
									vy=readfloat f
									vz=readfloat f
									append (objDataArr[globObjCount]) [-vx,vz,vy]
								)
							)
							default:
							(
								messageBox "Unknown VertexArray type. PM me."
							)

						)
					)
					21:	-- vertex buffer
					(
						fseek f 12#seek_cur

						defaultColor = readlong f
						positions    = readlong f -- ObjectIndex

						mPos   = [readfloat f,readfloat f,readfloat f]
						mScale = [1,1,1]*readfloat f

						normals = readlong f -- ObjectIndex
						colors  = readlong f -- ObjectIndex

						tArr=#()
						tArrCount = readlong f
						for i = 1 to tArrCount do
						(
							texCoords = readlong f
							tPos   = [readfloat f,readfloat f,readfloat f]
							tScale = [1,1,1]*readfloat f

							append tArr (objDataArr[texCoords])
						)

						vArr = objDataArr[positions]
						nArr = objDataArr[normals]
					)
					11: -- triangle strip array
					(
						fArr=#()
						fseek f 21#seek_cur

						stripType = readbyte f#unsigned
						count = readlong f#unsigned

						case stripType of
						(
							0x80: -- UInt32, haven't found this in RR3 meshes yet but included it just in case
							(
								for i = 1 to count/3 do
								(
									fa=readlong f#unsigned
									fb=readlong f#unsigned
									fc=readlong f#unsigned
									append fArr([fa,fb,fc]+1)
								)
							)
							0x81: -- Byte
							(
								for i = 1 to count/3 do
								(
									fa=readbyte f#unsigned
									fb=readbyte f#unsigned
									fc=readbyte f#unsigned
									append fArr([fa,fb,fc]+1)
								)
							)
							0x82: -- UInt16
							(
								for i = 1 to count/3 do
								(
									fa=readshort f#unsigned
									fb=readshort f#unsigned
									fc=readshort f#unsigned
									append fArr([fa,fb,fc]+1)
								)
							)
							default:
							(
								messageBox "Unknown strip type. PM me."
							)
						)
						long1 = readlong f
						long2 = readlong f
					)
					14: -- Loader (not serializable)
					(
						basePosition=ftell f
						fseek f 68#seek_cur
						objectMatID = 1+readlong f
						mshName = readstring f

						fseek f (basePosition+objectLength) #seek_set

						with redraw off
						(
							if vArr.count > 0 do
							(
								msh = mesh vertices:vArr faces:fArr name:mshName

								matchingMaterialFound = false
								for m in mArr do
								(
									if m.name == objectMatID as string do
									(
										matchingMaterialFound = true
										msh.material = m
									)
								)

								if not matchingMaterialFound do
								(
									aMaterial = standard showInViewport:true Name:(objectMatID as string)
									append mArr aMaterial
									msh.material = aMaterial
								)

								if nArr.count == vArr.count do
								(
									for j = 1 to fArr.count do setFaceSmoothGroup msh j 1
									select msh
									addmodifier msh (Edit_Normals ()) ui:off
									msh.Edit_Normals.MakeExplicit selection:#{1..nArr.count}
									EN_convertVS = msh.Edit_Normals.ConvertVertexSelection
									EN_setNormal = msh.Edit_Normals.SetNormal
									normID = #{}
									--apply normals
									for k = 1 to vArr.count do
									(
										free normID
										EN_convertVS #{k} &normID
										for id in normID do EN_setNormal id nArr[k]
									)
									collapseStack msh
								)

								for i = 1 to tArr.count do
								(
									meshop.setNumMaps msh (i+1) keep:true
									meshop.setMapSupport msh i true
									for v = 1 to tArr[i].count do meshop.setMapVert msh i v tArr[i][v]
								)

								msh.transform = (scaleMatrix mScale) * (transMatrix mPos)
								--if (findstring mshName "DAMAGE" != undefined) or (findstring mshName "BROKEN" != undefined) do hide msh
							)
						)

						free fArr
					)	

					24: -- material list
					(
						MatName=#()
						materialCount = readlong f
						for i = 1 to materialCount do append MatName (readstring f)
						for i = 1 to mArr.count do
						(
							mArr[i].name = MatName[mArr[i].name as integer]
							if i <= meditmaterials.count do meditmaterials[i] = mArr[i]
						)
						free MatName
					)
					default:
					(
						fseek f objectLength#seek_cur
					)
				)
			)
			sectionChecksum = readlong f
		)
		gc()
		fclose f 

		max select none
		--hide helpers
		cui.expertModeOff()
		messageBox "Import complete!"
	)
)